/**************
* Program:	DV Icy (Dreamcast VMU Icon Viewer and Converter)
* Author:	Sean Robinson
* License:	GNU Affero General Public License Version 3 (or later)
* Released:	2019-06-24
* Version:	0.8
**************/

// Note to self: Always upgrade the ver. number in both the preamble and in printVerInfo().

//#ifdef _WIN32
//	#include <conio.h>
//#endif

#include <fstream>
#include <iostream>
#include <string>
#include "SDL/SDL.h"
using namespace std;

/*
void winPause()
{
#ifdef _WIN32
	cout << "Press any key to continue...";
	_getch();
	cin.clear();
#endif
}
*/


void convertBWiconToArray(ifstream &vmuFile, bool bwIcon[])
{
	// Find black-and-clear icon starting offset
	vmuFile.seekg(16);
	vmuFile.seekg(vmuFile.get());

	// Translate from hexadecimal to pixels
	int pixelValues = vmuFile.get();
	int place = 0;
	for(int j = 0; j < 128; j++)
	{
		for(int i = 128; i > 0; i /= 2)
		{
			// Store icon pixels as trues and falses
			if(pixelValues >= i)
			{
				bwIcon[place] = true;
				pixelValues -= i;
			}
			else
			{
				bwIcon[place] = false;
			}
			place++;
		}
		pixelValues = vmuFile.get();
	}
}


void convertCLiconToArray(ifstream &vmuFile, int clIcon[])
{
	// Find color icon starting offset
	vmuFile.seekg(20);
	if(vmuFile.peek() == 0)
	{
		cout << "No color icon found in VMU file." << endl;
		// If no color icon, fill color data with 0's
		for(int k = 0; k < 1088; k++)
		{
			clIcon[k] = 0;
		}
		return;
	}
	vmuFile.seekg(vmuFile.get());

	// Translate from hexadecimal to 16 colors, 4 channels per color (16xGBAR=64)
	int pixelValues;
	for(int k = 0; k < 16; k++)
	{
		pixelValues = vmuFile.get();
		// Convert high-nybble to green channel, 0-15
		clIcon[k*4+0] = pixelValues / 16;
		// Convert low-nybble to blue channel, 0-15
		clIcon[k*4+1] = pixelValues % 16;
		pixelValues = vmuFile.get();
		// Convert high-nybble to alpha channel, 0-15
		clIcon[k*4+2] = pixelValues / 16;
		// Convert low-nybble to red channel, 0-15
		clIcon[k*4+3] = pixelValues % 16;
	}

	// Translate from hexadecimal to colors 0-15 for 32x32 pixels (1024)
	for(int k = 64; k < 1088; k += 2)
	{
		pixelValues = vmuFile.get();
		// Convert high-nybble to first color
		clIcon[k] = pixelValues / 16;
		// Convert low-nybble to second color
		clIcon[k+1] = pixelValues % 16;
	}
}


void displayBWIcon(bool bwIcon[])
{
	SDL_Surface *screen;

	/* Have a preference for 8-bit, but accept any depth */
	screen = SDL_SetVideoMode(320, 320, 8, SDL_SWSURFACE|SDL_ANYFORMAT);
	if ( screen == NULL ) {
		fprintf(stderr, "Couldn't set 640x480x8 video mode: %s\n",
						SDL_GetError());
		exit(1);
	}
	printf("Set 640x480 at %d bits-per-pixel mode\n",
		screen->format->BitsPerPixel);

	/* Code to set the screen to be yellow */

	Uint32 bgColor;

	/* Map the color yellow to this display (R=0xff, G=0xFF, B=0x00)
		Note: If the display is palettized, you must set the palette first.
	*/
	bgColor = SDL_MapRGB(screen->format, 0xff, 0x00, 0xff);
	Uint32 fgColor = SDL_MapRGB(screen->format, 0x00, 0x00, 0x00);

	/* Lock the screen for direct access to the pixels */
	if ( SDL_MUSTLOCK(screen) ) {
		if ( SDL_LockSurface(screen) < 0 ) {
			fprintf(stderr, "Can't lock screen: %s\n", SDL_GetError());
			exit(1);
		}
	}

	SDL_FillRect(screen, NULL, bgColor);

	// Complete the VMU icon's pattern
	SDL_Rect blockOfColor;
	int i = 0, k = 0;
	for(int j = 0; j < 1024; j++)
	{
		if(bwIcon[j])
		{
			blockOfColor.x = 10 * k;
			blockOfColor.y = 10 * i;
			blockOfColor.w = 10;
			blockOfColor.h = 10;
			SDL_FillRect(screen, &blockOfColor, fgColor);
		}
		k++;

		if((j + 1) % 32 == 0)
		{
			i++;
			k = 0;
		}
	}

	if ( SDL_MUSTLOCK(screen) ) {
		SDL_UnlockSurface(screen);
	}
	/* Update just the part of the display that we've changed */
	SDL_Flip(screen);
	printf("Quitting SDL.\n");

	SDL_Delay(1000);

	/* Shutdown all subsystems */
	SDL_Quit();
}


void displayCLIcon(int clIcon[])
{
	SDL_Surface *screen;

	/* Have a preference for 8-bit, but accept any depth */
	screen = SDL_SetVideoMode(320, 320, 8, SDL_SWSURFACE|SDL_ANYFORMAT);
	if ( screen == NULL ) {
		fprintf(stderr, "Couldn't set 640x480x8 video mode: %s\n",
						SDL_GetError());
		exit(1);
	}
	printf("Set 640x480 at %d bits-per-pixel mode\n",
		screen->format->BitsPerPixel);

	Uint32 bgColor = SDL_MapRGB(screen->format, 0xff, 0x00, 0xff); // magenta

	/* Lock the screen for direct access to the pixels */
	if ( SDL_MUSTLOCK(screen) ) {
		if ( SDL_LockSurface(screen) < 0 ) {
			fprintf(stderr, "Can't lock screen: %s\n", SDL_GetError());
			exit(1);
		}
	}

	SDL_FillRect(screen, NULL, bgColor);

	// Complete the VMU icon's pattern
	SDL_Rect blockOfColor;
	blockOfColor.w = 10;
	blockOfColor.h = 10;
	int i = 0, k = 0;
	for(int j = 64; j < 1088; j++)
	{
		blockOfColor.x = 10 * k;
		blockOfColor.y = 10 * i;
		// Change color map from GBAR->RGBA
		Uint32 fgColor = SDL_MapRGBA(screen->format,
			clIcon[clIcon[j]*4+3]*17, clIcon[clIcon[j]*4+0]*17,
			clIcon[clIcon[j]*4+1]*17, clIcon[clIcon[j]*4+2]*17);
		SDL_FillRect(screen, &blockOfColor, fgColor);
		k++;

		if((j + 1) % 32 == 0)
		{
			i++;
			k = 0;
		}
	}

	if ( SDL_MUSTLOCK(screen) ) {
		SDL_UnlockSurface(screen);
	}
	/* Update just the part of the display that we've changed */
	SDL_Flip(screen);
	printf("Quitting SDL.\n");

	SDL_Delay(1000);

	/* Shutdown all subsystems */
	SDL_Quit();
}


void printBWIcon(bool bwIcon[])
{
	// Use the "full block" character (decimal code 219, hexadecimal code 0x00DB) for the printed characters
	char print_char = 219;
	// Display icon as X's and white spaces
	for(int j = 0; j < 1024; j++)
	{
		if(bwIcon[j])
		{
			cout << print_char;
		}
		else
		{
			cout << " ";
		}

		if((j + 1) % 32 == 0)
		{
			cout << endl;
		}
	}
}


void printCLIcon(int clIcon[])
{
	// Use the "full block" character (decimal code 219, hexadecimal code 0x00DB) for the printed characters
	char print_char = 219;
	// Display icon as print_char's and white spaces
	for(int j = 64; j < 1088; j++)
	{
		if(clIcon[j])
		{
			cout << print_char;
		}
		else
		{
			cout << " ";
		}

		if((j + 1) % 32 == 0)
		{
			cout << endl;
		}
	}
}


void printHelpInfo()
{
	// Print help and program options
	cout << "Usage: dv_icy-dreamcast_vmu_icon_viewer_and_converter [options] [file]" << endl;
	
	cout << "\t" << "-b, --printb";
	cout << "\t" << "Print the black-and-clear icon to standard output" << endl;

	cout << "\t" << "-B, --displayb";
	cout << "\t" << "Display the black-and-clear icon on the screen" << endl;

	cout << "\t" << "-c, --printc";
	cout << "\t" << "Print the color icon (in black-and-white) to standard" << endl;
	cout << "\t" << "\t" << "\t" << "output" << endl;

	cout << "\t" << "-C, --displayc";
	cout << "\t" << "Display the color icon on the screen" << endl;

	cout << "\t" << "-h, --help";
	cout << "\t" << "Print this help message" << endl;

	cout << "\t" << "-v, --version";
	cout << "\t" << "Print the version number" << endl;
}


void printVerInfo()
{
	const float version = 0.8;
	// Print version info
	cout << "DV Icy (Dreamcast VMU Icon Viewer and Converter)" << " ";
	cout << version << endl;
	cout << "Copyright (C) 2019 Nz17 Productions" << endl;
	cout << "This is free software; see the source for copying conditions." << endl;
	cout << "There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE." << endl;
}


int main(int argc, char *argv[])
{
	// Try to reclaim standard output and standard error from SDL (on Windows)
	#ifdef _WIN32
		freopen( "CON", "w", stdout);
		freopen( "CON", "w", stderr);
	#endif

	bool help = false, version = false;
	bool displayBW = false, displayCL = false;
	bool printBW = false, printCL = false;
	string vmuFileName = "";

	// Parse command line arguments
	for(int i = 1; i < argc; i++)
	{
		if(strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-h") == 0)
		{
			help = true;
		}
		else if(strcmp(argv[i], "--version") == 0 || strcmp(argv[i], "-v") == 0)
		{
			version = true;
		}
		else if(strcmp(argv[i], "--displayb") == 0 || strcmp(argv[i], "-B") == 0)
		{
			displayBW = true;
		}
		else if(strcmp(argv[i], "--displayc") == 0 || strcmp(argv[i], "-C") == 0)
		{
			displayCL = true;
		}
		else if(strcmp(argv[i], "--printb") == 0 || strcmp(argv[i], "-b") == 0)
		{
			printBW = true;
		}
		else if(strcmp(argv[i], "--printc") == 0 || strcmp(argv[i], "-c") == 0)
		{
			printCL = true;
		}
		else
		{
			vmuFileName = argv[i];
		}
	}

	// Print version info
	if(version)
	{
		printVerInfo();
	}

	// Print help info
	if(help)
	{
		printHelpInfo();
	}

	// If there is no file name argument, then print the help screen and exit
	if(vmuFileName == "" && help != true && version != true)
	{
		printHelpInfo();
	}

	// If there is a file name argument, then open the VMU's icon
	if(vmuFileName != "")
	{
		// Read VMU file
		ifstream vmuFile(vmuFileName, ifstream::binary);
		if(vmuFile.good())
			cout << "Successfully read \"" << vmuFileName << "\"." << endl << endl;
		else
		{
			cerr << "Invalid file name or file not found: \"" << vmuFileName << "\".";
			cerr << endl << endl;

			return 1;
		}

		// Arrays for the icon name, the black-and-clear icon, and the color icon
		char iconName [17];
		bool bwIcon [1024];
		int clIcon [1088];

		// Convert icon description to a character array
		for(int i = 0; i < 16; i++)
			iconName[i] = (char) vmuFile.get();
		iconName[16] = '\0';

		// If there was a file name argument and the display mode was not
		// specified, then default to displayCL
		if(!displayBW && !displayCL && !printBW && !printCL)
		{
			displayCL = true;
		}

		// Convert black-and-clear icon to a boolean array
		if(displayBW || printBW)
		{
			convertBWiconToArray(vmuFile, bwIcon);
		}

		// Convert color icon to a boolean array
		if(displayCL || printCL)
		{
			convertCLiconToArray(vmuFile, clIcon);
		}

		// Close VMU file
		vmuFile.close();

		// Print the file's internal name, comment, or description
		cout << "The icon's name, comment, or description:" << endl;
		cout << iconName << endl << endl;

		// Display the file's icon on the screen's display
		if(displayBW || displayCL)
		{
			// Initialize SDL
			if(SDL_Init(SDL_INIT_VIDEO) < 0)
			{
				cerr << "Unable to initialize SDL: " << SDL_GetError() << endl;
				return 1;
			}
			atexit(SDL_Quit);

			if(displayBW)
			{
				displayBWIcon(bwIcon);
			}
			// If there is a file name argument but no other arguments specified, then open the VMU's icon in "display color icon" mode
			else
			{
				displayCLIcon(clIcon);
			}
		}

		// Print the file's icon on the terminal / console
		// Print the black-and-clear icon to standard output
		if(printBW)
		{
			cout << "Black-and-clear icon:" << endl;
			printBWIcon(bwIcon);
		}
		// Print the color icon to standard output
		else if(printCL)
		{
			cout << "Color icon:" << endl;
			// Fix: Add a function to convert color icons (clIcon) to black-and-white icons (bwIcon) for printing
			printCLIcon(clIcon);
		}
		cout << endl;
	}

	return 0;
}
